#encoding=utf8
from django.conf import settings
from django.core.exceptions import (ObjectDoesNotExist,
                                    ImproperlyConfigured,
)
from django_load.core import load_object


def get_test_user_function():
    """
    Returns the function set on settings.CHATROOMS_TEST_USER_FUNCTION
    if any, else returns False

    """
    if hasattr(settings, 'CHATROOMS_TEST_USER_FUNCTION'):
        try:
            return load_object(settings.CHATROOMS_TEST_USER_FUNCTION)
        except (ImportError, TypeError):
            raise ImproperlyConfigured(
                "The variable set as settings.CHATROOMS_TEST_USER_FUNCTION "
                "is not a module or the path is not correct"
            )
    return False

test_user_function = get_test_user_function()


def check_user_passes_test(request, user):
    """
    Returns the test_user_function if any,
    else returns True

    """
    if test_user_function:
        return test_user_function(request, user)
    return True



# encoding: utf-8

import json
import itertools
from datetime import datetime, timedelta
from collections import deque

from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.http import (HttpResponse,
                        HttpResponseBadRequest)
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator

from gevent.event import Event

from ..models import Room
from ..signals import chat_message_received
from ..utils.auth import check_user_passes_test
from ..utils.decorators import ajax_user_passes_test_or_403
from ..utils.decorators import ajax_login_required
from ..utils.handlers import MessageHandlerFactory


TIME_FORMAT = '%Y-%m-%dT%H:%M:%S:%f'

TIMEOUT = 30
if settings.DEBUG:
    TIMEOUT = 3


class ChatView(object):
    def __init__(self):
        """
        Defines dictionary attibutes sorted by room_id
        For each room:
        - new_message_events contains gevent.Event objects used by message
          handlers to pause/restart execution and implement long polling
        - messages stores the queue of the latest 50 messages
        - counters contains iterators to pick up message identifiers
        - connected_users is a dictionary holding the usernames of connected
          users sorted by the time of their latest request
        - new_connected_user_event contains gevent.Event objects used
          by self.notify_users_list and self.get_users_list methods to
          implement long polling

        """
        self.new_message_events = {}
        self.messages = {}
        self.counters = {}
        self.connected_users = {}
        self.new_connected_user_event = {}
        rooms = Room.objects.all()
        for room in rooms:
            self.new_message_events[room.id] = Event()
            self.messages[room.id] = deque(maxlen=50)
            self.counters[room.id] = itertools.count()
            self.connected_users[room.id] = {}
            self.new_connected_user_event[room.id] = Event()

    def signal_new_message_event(self, room_id):
        """Signals new_message_event given a room_id """
        self.new_message_events[room_id].set()
        self.new_message_events[room_id].clear()

    def wait_for_new_message(self, room_id, timeout=TIMEOUT):
        """Waits for new_message_event given a room_id """
        self.new_message_events[room_id].wait(timeout)

    def get_messages_queue(self, room_id):
        """Returns the message queue given a room_id """
        return self.messages[room_id]

    def get_next_message_id(self, room_id):
        """Returns the next message identifier given a room_id """
        return self.counters[room_id].next()

    def get_connected_users(self, room_id):
        """Returns the connected users given a room_id"""
        return self.connected_users[room_id]

    @method_decorator(ajax_login_required)
    @method_decorator(ajax_user_passes_test_or_403(check_user_passes_test))
    def get_messages(self, request):
        """Handles ajax requests for messages
        Requests must contain room_id and latest_id
        Delegates MessageHandler.retrieve_message method to return the list
        of messages

        """
        try:
            room_id = int(request.GET['room_id'])
            latest_msg_id = int(request.GET['latest_message_id'])
        except:
            return HttpResponseBadRequest(
            "Parameters missing or bad parameters. "
            "Expected a GET request with 'room_id' and 'latest_message_id' "
            "parameters")
        get_object_or_404(Room, id=room_id)

        handler = MessageHandlerFactory()
        messages = handler.retrieve_messages(self, room_id)

        to_jsonify = [
            {"message_id": msg_id,
             "username": message.user.username,
             "date": message.date.strftime(TIME_FORMAT),
             "content": message.content}
            for msg_id, message in messages
            if msg_id > latest_msg_id
        ]
        return HttpResponse(json.dumps(to_jsonify),
                            mimetype="application/json")

    @method_decorator(ajax_login_required)
    @method_decorator(ajax_user_passes_test_or_403(check_user_passes_test))
    def send_message(self, request):
        """Gets room_id and message from request and sends a
        chat_message_received signal

        """
        try:
            room_id = int(request.POST['room_id'])
            message = request.POST['message']
            date = datetime.now()
        except:
            return HttpResponseBadRequest(
            "Parameters missing or bad parameters"
            "Expected a POST request with 'room_id' and 'message' parameters")
        get_object_or_404(Room, id=room_id)
        user = request.user
        foo, response = chat_message_received.send(
                            sender=self,
                            room_id=room_id,
                            user=user,
                            message=message,
                            date=date)[0]

        return HttpResponse(json.dumps({
                 'timestamp': date.strftime(TIME_FORMAT), }
        ))

    @method_decorator(ajax_login_required)
    @method_decorator(ajax_user_passes_test_or_403(check_user_passes_test))
    def notify_users_list(self, request):
        """Updates user time into connected users dictionary """
        try:
            room_id = int(request.POST['room_id'])
        except:
            return HttpResponseBadRequest(
            "Parameters missing or bad parameters"
            "Expected a POST request with 'room_id'")

        user = request.user
        get_object_or_404(Room, id=room_id)
        date = datetime.today()
        self.connected_users[room_id].update({user.username: date})
        self.new_connected_user_event[room_id].set()
        self.new_connected_user_event[room_id].clear()
        return HttpResponse('Connected')

    @method_decorator(ajax_login_required)
    @method_decorator(ajax_user_passes_test_or_403(check_user_passes_test))
    def get_users_list(self, request):
        """Dumps the list of connected users """
        REFRESH_TIME = 8
        try:
            room_id = int(request.GET['room_id'])
        except:
            return HttpResponseBadRequest(
            "Parameters missing or bad parameters"
            "Expected a POST request with 'room_id'")

        get_object_or_404(Room, id=room_id)
        user = request.user
        self.connected_users[room_id].update({
                                user.username: datetime.today()
                            })
        self.new_connected_user_event[room_id].wait(REFRESH_TIME)

        # clean connected_users dictionary of disconnected users
        self._clean_connected_users(room_id)
        json_users = [
            {"username": _user,
             "date": _date.strftime(TIME_FORMAT)}
            for _user, _date in self.connected_users[room_id].iteritems()
        ]
        json_response = {
            "now": datetime.today().strftime(TIME_FORMAT),
            "users": json_users,
            "refresh": str(REFRESH_TIME),
        }
        return HttpResponse(json.dumps(json_response),
                            mimetype='application/json')

    @method_decorator(ajax_login_required)
    @method_decorator(ajax_user_passes_test_or_403(check_user_passes_test))
    def get_latest_message_id(self, request):
        """Dumps the id of the latest message sent """
        try:
            room_id = int(request.GET['room_id'])
        except:
            return HttpResponseBadRequest()
        latest_msg_id = -1
        msgs_queue = self.messages[room_id]
        if msgs_queue:
            latest_msg_id = msgs_queue[-1][0]
        response = {"id": latest_msg_id}
        return HttpResponse(json.dumps(response), mimetype="application/json")

    def _clean_connected_users(self, room_id, seconds=60):
        """Remove from connected users dictionary users not seen
        for seconds

        """
        now = datetime.today()
        for usr, date in self.connected_users[room_id].items():
            if (now - timedelta(seconds=seconds)) > date:
                self.connected_users[room_id].pop(usr)


chat = ChatView()
send_message = chat.send_message
get_messages = chat.get_messages
get_users_list = chat.get_users_list
notify_users_list = chat.notify_users_list
get_latest_message_id = chat.get_latest_message_id


@receiver(post_save, sender=Room)
def create_events_for_new_room(sender, **kwargs):
    """Creates an entry in Chat dictionary attributes
    when a new room is created

    """
    if kwargs.get('created'):
        instance = kwargs.get('instance')
        room_id = instance.id
        chat.new_message_events[room_id] = Event()
        chat.messages[room_id] = deque(maxlen=50)
        chat.counters[room_id] = itertools.count()
        chat.connected_users[room_id] = {}
        chat.new_connected_user_event[room_id] = Event()



#encoding=utf8

from django.conf.urls.defaults import *
from django.contrib.auth.decorators import login_required

from . import views

urlpatterns = patterns('chatrooms',
    # room views
    url(r'^rooms/$',
        login_required(views.RoomsListView.as_view()),
        name="rooms_list"),
    url(r'^room/(?P<slug>[-\w\d]+)/$',
        login_required(views.RoomView.as_view()),
        name="room_view"),

    # ajax requests
    url(r'^get_messages/', 'ajax.chat.get_messages'),
    url(r'^send_message/', 'ajax.chat.send_message'),
    url(r'^get_latest_msg_id/', 'ajax.chat.get_latest_message_id'),
    url(r'^get_users_list/$', 'ajax.chat.get_users_list'),
    url(r'^notify_users_list/$', 'ajax.chat.notify_users_list')
)



#encoding=utf8

from django.db import models
from django.contrib.auth.models import User

from polymorphic import PolymorphicModel


class Room(PolymorphicModel):
    name = models.CharField(max_length=64, unique=True)
    slug = models.SlugField()
    subscribers = models.ManyToManyField(User, blank=True)
    private = models.NullBooleanField()
    password = models.CharField(max_length=32, blank=True)

    def __unicode__(self):
        return u"%s" % self.name

    @models.permalink
    def get_absolute_url(self):
        return ('room_view', [self.slug])


class Message(PolymorphicModel):
    user = models.ForeignKey(User)
    date = models.DateTimeField(['%Y-%m-%d %H:%M:%S:%f'])
    room = models.ForeignKey(Room)
    content = models.CharField(max_length=5000)



from utils.handlers import MessageHandlerFactory
from django.dispatch import Signal


chat_message_received = Signal(
    providing_args=[
        "room_id",
        "user",
        "message",
        "date",
])

handler = MessageHandlerFactory()

chat_message_received.connect(handler.handle_received_message)



#encoding=utf8

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.decorators import available_attrs
from django.utils.functional import wraps


def ajax_user_passes_test_or_403(test_func, message="Access denied"):
    """
    Decorator for views that checks the user passes the given test,
    raising 403 if user does not pass test.
    If the request is ajax returns a 403 response with a message,
    else renders a 403.html template.
    The test should be a callable which takes the user object and
    returns True if the user passes.
    """

    def decorator(view_func):
        @wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(request, *args, **kwargs):
            if test_func(request, request.user):
                return view_func(request, *args, **kwargs)
            # returns an HttpResponseForbidden if request is ajax
            if request.is_ajax:
                return HttpResponseForbidden(message)
            ctx = RequestContext(request)
            resp = render_to_response('403.html', context_instance=ctx)
            resp.status_code = 403
            return resp
        return _wrapped_view
    return decorator


def ajax_login_required(view_func):
    """Handle non-authenticated users differently if it is an AJAX request
    Borrowed from django-hilbert
    """

    @wraps(view_func, assigned=available_attrs(view_func))
    def _wrapped_view(request, *args, **kwargs):
        if request.is_ajax():
            if request.user.is_authenticated():
                return view_func(request, *args, **kwargs)
            else:
                response = HttpResponse()
                response['X-Django-Requires-Auth'] = True
                response['X-Django-Login-Url'] = settings.LOGIN_URL
                return response
        else:
            return login_required(view_func)(request, *args, **kwargs)
    return _wrapped_view


def signals_new_message_at_end(func):
    """Decorator for MessageHandler.handle_received_message method
    """

    @wraps(func, assigned=available_attrs(func))
    def _wrapper(self, sender, room_id, user, message, date, **kwargs):
        f = func(self, sender, room_id, user, message, date, **kwargs)
        sender.signal_new_message_event(room_id)
        return f
    return _wrapper


def waits_for_new_message_at_start(func):
    """Decorator for MessageHandler.retrieve_messages method
    """
    @wraps(func, assigned=available_attrs(func))
    def _wrapper(self, chatobj, room_id, *args, **kwargs):
        chatobj.wait_for_new_message(room_id)
        return func(self, chatobj, room_id, *args, **kwargs)
    return _wrapper



#encoding=utf8

from django.contrib import admin
from models import Room, Message


class RoomAdmin(admin.ModelAdmin):
    prepopulated_fields = {"slug": ("name",)}


class MessageAdmin(admin.ModelAdmin):
    pass


admin.site.register(Room, RoomAdmin)
admin.site.register(Message, MessageAdmin)



import json

from django.test import TestCase
from django.test.client import Client

from django.contrib.auth.models import User

from chatrooms.models import Room


class SimpleTest(TestCase):
    def create_user(self):
        # creates a user
        username = 'john'
        password = 'johnpasswd'
        email = 'john@beatles.com'
        user = User.objects.create_user(username=username,
                                        password=password,
                                        email=email)
        user.save()
        return username, password

    def test_get_messages(self, *args, **kwargs):
        username, password = self.create_user()

        # login user
        client = Client()
        client.login(username=username, password=password)

        # creates a room
        room = Room()
        room.save()

        # message queue empty: check last_message_id
        response = client.get('/chat/get_last_msg_id/?room_id=%d' % room.id)
        json_response = json.loads(response.content)
        last_msg_id = json_response['id']
        self.assertEquals(last_msg_id, -1)

        # posts a message
        post_response = client.post('/chat/send_message/',
                                {'room_id': room.pk,
                                 'message': 'ABCD'},
                HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEquals(post_response.status_code, 200)
        json_response = json.loads(post_response.content)
        timestamp = json_response['timestamp']

        # gets list of messages
        response = client.get('/chat/get_messages/?room_id=%d&latest_id=%d' % (
                                                        room.id, last_msg_id),
                HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        json_response = json.loads(response.content)

        expected_json = [{u'message_id': 0,
                          u'username': u'john',
                          u'date': timestamp,
                          u'content': u'ABCD', }]
        self.assertEquals(expected_json, json_response)

        # check last_message_id
        response = client.get('/chat/get_last_msg_id/?room_id=%d' % room.id)
        json_response = json.loads(response.content)
        last_msg_id = json_response['id']
        self.assertEquals(last_msg_id, 0)



#encoding=utf8

from django.template import RequestContext
from django.views.generic import ListView, DetailView

from .models import Room


class RoomsListView(ListView):
    """View to show the list of rooms available """
    context_object_name = "rooms"
    queryset = Room.objects.all()
    template_name = "chatrooms/rooms_list.html"
    paginate_by = 2


class RoomView(DetailView):
    """View for the single room """
    model = Room
    context_object_name = 'room'
    template_name = "chatrooms/room.html"



from setuptools import setup, find_packages
import os

version = '0.1'

setup(name='django-chatrooms',
      version=__import__('chatrooms').__version__,
      description="A django app providing reverse-ajax chat rooms",
      long_description=open("README.rst").read(),
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        ],
      keywords='django chat ajax',
      author='Vincenzo E. Antignano (@qubird)',
      author_email='',
      url='https://github.com/qubird/django-chatrooms',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      include_package_data=True,
      namespace_packages=[],
      zip_safe=False,
      install_requires=[
          'setuptools',
          'gevent',
          'django_load>=1.0',
          'django_polymorphic>=0.2',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



from django.core.exceptions import ObjectDoesNotExist


def check_user_is_subscribed(request, user):
    """
    Example of function settable as settings.CHATROOMS_TEST_USER_FUNCTION
    Takes request and user as arguments
    Returns True if the request user is subscribed to the request chat room,
    False otherwise

    """
    room_id = request.GET['room_id']
    try:
        user.room_set.get(pk=room_id)
        return True
    except ObjectDoesNotExist:
        return False



#encoding=utf8

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django_load.core import load_object

from .decorators import (signals_new_message_at_end,
                        waits_for_new_message_at_start)
from ..models import Room, Message


class MessageHandler(object):
    """
    Class which implements two methods:
    - handle_received_message
    is designated to handle the "chat_message_received" signal
    - retrieve_messages
    is designated to return the list of messages sent to chat room so far

    ``handle_received_message`` method is designed to perform operations
    with the received message such that ``retrieve_messages`` is able to
    retrieve it afterwards.

    These methods are responsible for long polling implementation:
    ``retrieve_messages`` waits for new_message_event at its start,
    ``handle_received_message`` signals new_message_event at its end.

    The handlers can be customized and replaced extending this class
    and setting the full path name of the extending class
    into settings.CHATROOMS_HANDLERS_CLASS
    """

    @signals_new_message_at_end
    def handle_received_message(self,
        sender, room_id, user, message, date, **kwargs):
        """
        Default handler for the message_received signal.
        1 - Saves an instance of message to db
        2 - Appends a tuple (message_id, message_obj)
            to the sender.messages queue
        3 - Signals the "New message" event on the sender (decorator)
        4 - Returns the created message

        """
        # 1
        room = Room.objects.get(id=room_id)
        new_message = Message(user=user,
                              room=room,
                              date=date,
                              content=message)
        new_message.save()

        # 2
        msg_number = sender.get_next_message_id(room_id)
        messages_queue = sender.get_messages_queue(room_id)
        messages_queue.append((msg_number, new_message))

        # 3 - decorator does
        # sender.signal_new_message_event(room_id)

        # 4
        return new_message

    @waits_for_new_message_at_start
    def retrieve_messages(self, chatobj, room_id, *args, **kwargs):
        """
        Returns a list of tuples like:
        [(message_id, message_obj), ...]
        Where message_obj is an instance of Message or an object with
        the attributes 'username', 'date' and 'content' at least

        1 - Waits for new_message_event (decorator)
        2 - returns the queue of messages stored in
        the ChatView.message dictionary by self.handle_received_message

        """
        # 1 - decorator does
        # chatobj.wait_for_new_message(room_id)

        # 2
        return chatobj.get_messages_queue(room_id)


class MessageHandlerFactory(object):
    """
    Returns a (singleton) instance of the class set as
    settings.CHATROOMS_HANDLERS_CLASS
    if any, else returns an instance of MessageHandler

    """
    _instance = None

    def __new__(cls):
        klass = MessageHandler
        if hasattr(settings, 'CHATROOMS_HANDLERS_CLASS'):
            try:
                klass = load_object(settings.CHATROOMS_HANDLERS_CLASS)
            except (ImportError, TypeError):
                raise ImproperlyConfigured(
                    "The class set as settings.CHATROOMS_HANDLERS_CLASS "
                    "does not exists"
                )

        if not cls._instance:
            cls._instance = klass()
        return cls._instance



from gevent import monkey
from gevent import __version__ as gevent_version
monkey.patch_all()

from django.core.management.commands import test


class Command(test.Command):
    def handle(self, *args, **kw):
        print "using GEvent %s" % gevent_version
        super(Command, self).handle(*args, **kw)
